from aws_cdk.aws_cloudwatch import CfnAlarm
from aws_cdk.aws_lambda import IFunction
from aws_cdk.aws_sns import ITopic
from aws_cdk.core import Stack


class LambdaAlarms:
    def __init__(
            self,
            scope: Stack,
            prefix: str,
            sns_topic: ITopic,
            lambda_function: IFunction
    ) -> None:
        self.__prefix = prefix
        self.__scope = scope
        self.__sns_topic = sns_topic
        self.__lambda_function = lambda_function

        self.__errors_alarm = CfnAlarm(
            scope=self.__scope,
            id=self.__prefix + 'ErrorsCountAlarm',
            actions_enabled=True,
            alarm_name=self.__prefix + 'ErrorsCountAlarm',
            alarm_description=f'Lambda function {self.__lambda_function.function_name} errors count alarm.',
            dimensions=[CfnAlarm.DimensionProperty(name='FunctionName', value=self.__lambda_function.function_name)],
            metric_name='Errors',
            namespace='AWS/Lambda',
            alarm_actions=[
                self.__sns_topic.topic_arn
            ],

            # Fire an alarm if at least one event has occurred within 10 minutes.
            # This way we will immanently know when a function has crashed or is crashing.
            evaluation_periods=1,
            period=600,
            comparison_operator='GreaterThanThreshold',
            threshold=0,
            statistic='Sum',
        )

        self.__throttles_alarm = CfnAlarm(
            scope=self.__scope,
            id=self.__prefix + 'ThrottlesAlarm',
            actions_enabled=True,
            alarm_name=self.__prefix + 'ThrottlesAlarm',
            alarm_description=f'Lambda function {self.__lambda_function.function_name} throttles count alarm.',
            dimensions=[CfnAlarm.DimensionProperty(name='FunctionName', value=self.__lambda_function.function_name)],
            metric_name='Throttles',
            namespace='AWS/Lambda',
            alarm_actions=[
                self.__sns_topic.topic_arn
            ],

            # Fire an alarm if average events is higher than 0 within 1 minute.
            # This way we can constantly push alarm notifications that a throttling is happening.
            evaluation_periods=1,
            period=60,
            comparison_operator='GreaterThanThreshold',
            statistic='Average',
            threshold=0,
        )

    @property
    def errors_alarm(self) -> CfnAlarm:
        return self.__errors_alarm

    @property
    def throttles_alarm(self) -> CfnAlarm:
        return self.__throttles_alarm



import re

from aws_ci_cd_lambda.lambda_alarms import LambdaAlarms
from aws_ci_cd_lambda.parameters.pipeline_parameters import PipelineParameters
from aws_ci_cd_lambda.parameters.lambda_parameters import LambdaParameters
from aws_ci_cd_lambda.parameters.vpc_parameters import VpcParameters
from aws_ci_cd_lambda.custom.initial_commit import InitialCommit
from aws_ci_cd_lambda.buildspec_object import BuildSpecObject
from aws_empty_bucket.empty_s3_bucket import EmptyS3Bucket
from aws_cdk import (
    aws_codepipeline_actions,
    aws_codepipeline,
    aws_codecommit,
    aws_codebuild,
    aws_lambda,
    aws_ec2,
    aws_iam,
    core
)


class CiCdLambda:
    def __init__(
            self,
            scope: core.Stack,
            prefix: str,
            pipeline_params: PipelineParameters,
            lambda_params: LambdaParameters,
            vpc_params: VpcParameters
    ):
        """
        AWS CDK package that helps deploying a lambda function.

        :param scope: A scope in which resources shall be created.
        :param prefix: Prefix for all of your resource IDs and names.
        :param pipeline_params: Parameters, letting you supply ssh key for accessing remote repositories.
        :param lambda_params: Parameters, focusing on the Lambda function itself.
        :param vpc_params: Parameters, focused on Virtual Private Cloud settings.
        """

        # CodeCommmit repository to store your function source code.
        self.project_repository = aws_codecommit.Repository(
            scope, prefix + 'CiCdLambdaCodeCommitRepo',
            repository_name=prefix + 'CiCdLambdaCodeCommitRepo',
        )

        # The lambda function for which this package is made.
        self.function = aws_lambda.Function(
            scope, prefix + 'Function',
            code=aws_lambda.Code.from_inline(
                'def runner():\n'
                '    return \'Hello, World!\''
            ),
            handler=lambda_params.lambda_handler,
            runtime=lambda_params.lambda_runtime,
            description=f'Lambda function {prefix}.',
            environment=lambda_params.environment,
            function_name=prefix,
            memory_size=lambda_params.lambda_memory,
            reserved_concurrent_executions=5,
            role=lambda_params.execution_role,
            security_groups=vpc_params.security_groups,
            timeout=core.Duration.seconds(lambda_params.lambda_timeout),
            vpc=vpc_params.vpc,
            vpc_subnets=aws_ec2.SubnetSelection(subnets=vpc_params.subnets)
        )

        # Create alarms for the function.
        if lambda_params.alarms_sns_topic:
            self.alarms = LambdaAlarms(scope, prefix, lambda_params.alarms_sns_topic, self.function)
        else:
            self.alarms = None

        # Convert bucket name to an S3 friendly one.
        bucket_name = self.__convert(prefix + 'CiCdLambdaArtifactsBucket')

        self.bucket = EmptyS3Bucket(
            scope, prefix + 'CiCdLambdaDeploymentBucket',
            bucket_name=bucket_name
        )

        # Create a BuildSpec object for CodeBuild
        self.buildspec = BuildSpecObject(
            prefix,
            self.bucket,
            pipeline_params.ssh_params.secret_id,
            pipeline_params.ssh_params.private_key,
            pipeline_params.install_args,
            pipeline_params.test_args,
            pipeline_params.custom_pre_build_commands
        )

        # CodeBuild project, that installs functions dependencies, runs tests and deploys it to Lambda.
        self.code_build_project = aws_codebuild.PipelineProject(
            scope, prefix + 'CiCdLambdaCodeBuildProject',
            project_name=prefix + 'CiCdLambdaCodeBuildProject',
            environment=aws_codebuild.BuildEnvironment(
                build_image=aws_codebuild.LinuxBuildImage.STANDARD_3_0,
                compute_type=aws_codebuild.ComputeType.SMALL,
                privileged=True
            ),
            build_spec=aws_codebuild.BuildSpec.from_object(self.buildspec.get_object()),
        )

        # Adding permissions that allow CodeBuild to do the aforementioned things.
        self.code_build_project.role.add_to_policy(
            statement=aws_iam.PolicyStatement(
                actions=[
                    's3:*',
                    'lambda:UpdateFunctionCode',
                ],
                resources=['*'],
                effect=aws_iam.Effect.ALLOW)
        )

        # If a secret is provided, we allow CodeBuild to read it.
        if pipeline_params.ssh_params.secret_arn is not None:
            self.code_build_project.role.add_to_policy(
                statement=aws_iam.PolicyStatement(
                    actions=[
                        'secretsmanager:GetSecretValue'
                    ],
                    resources=[pipeline_params.ssh_params.secret_arn],
                    effect=aws_iam.Effect.ALLOW)
            )

        # If KMS key is provided, we allow CodeBuild to decrypt using it.
        if pipeline_params.ssh_params.kms_key_arn is not None:
            self.code_build_project.role.add_to_policy(
                statement=aws_iam.PolicyStatement(
                    actions=[
                        "kms:Decrypt"
                    ],
                    effect=aws_iam.Effect.ALLOW,
                    resources=[pipeline_params.ssh_params.kms_key_arn]
                )
            )

        # Push hte initial commit to CodeCommit.
        self.initial_commit = InitialCommit(
            scope, prefix, self.project_repository
        ).get_resource()

        self.source_artifact = aws_codepipeline.Artifact(artifact_name=prefix + 'CiCdLambdaSourceArtifact')

        # CodePipeline source action to read from CodeCommit.
        self.source_action = aws_codepipeline_actions.CodeCommitSourceAction(
            repository=self.project_repository,
            branch='master',
            action_name='CodeCommitSource',
            run_order=1,
            trigger=aws_codepipeline_actions.CodeCommitTrigger.EVENTS,
            output=self.source_artifact
        )

        # CodePipeline build action that uses the CodeBuild project.
        self.build_action = aws_codepipeline_actions.CodeBuildAction(
            input=self.source_artifact,
            project=self.code_build_project,
            action_name='BuildAction',
            run_order=1
        )

        # CodePipeline pipeline that executes both actions.
        self.codecommit_to_lambda_pipeline = aws_codepipeline.Pipeline(
            scope,
            prefix + 'CiCdLambdaPipeline',
            pipeline_name=prefix + 'CiCdLambdaPipeline',
            artifact_bucket=self.bucket,
            stages=[
                aws_codepipeline.StageProps(
                    stage_name='SourceStage',
                    actions=[self.source_action]
                ),
                aws_codepipeline.StageProps(
                    stage_name='BuildStage',
                    actions=[self.build_action]
                )
            ]
        )

    @staticmethod
    def __convert(name: str) -> str:
        """
        Converts CamelCase string to pascal-case where underscores are dashes.
        This is required due to S3 not supporting capital letters or underscores.
        """
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1-\2', s1).lower()



from typing import Dict, Any, Optional
from aws_cdk.aws_iam import Role
from aws_cdk.aws_lambda import Runtime
from aws_cdk.aws_sns import ITopic


class LambdaParameters:
    """
    Parameters, focusing on the Lambda function itself.
    """
    def __init__(
            self,
            execution_role: Role,
            lambda_memory: int,
            lambda_timeout: int,
            lambda_handler: str,
            lambda_runtime: Runtime,
            environment: Optional[Dict[Any, Any]] = None,
            alarms_sns_topic: Optional[ITopic] = None
    ) -> None:
        """
        Constructor.

        :param execution_role: Execution role for your function.
        :param lambda_memory: Amount of memory your function can use, exceeding it will result in an error (in megabytes).
        :param lambda_timeout: Time, after which the function times out and returns an error (in seconds).
        :param lambda_handler: File name and function name that handles lambda invocation.
        :param lambda_runtime: Runtime for your Lambda function (the programming language it uses).
        """
        self.execution_role = execution_role
        self.lambda_memory = lambda_memory
        self.lambda_timeout = lambda_timeout
        self.lambda_handler = lambda_handler
        self.lambda_runtime = lambda_runtime
        self.environment: Dict[Any, Any] = environment or {}
        self.alarms_sns_topic = alarms_sns_topic



from typing import List
from aws_cdk.aws_ec2 import Vpc, Subnet, SecurityGroup


class VpcParameters:
    """
    Parameters, focused on Virtual Private Cloud settings.
    """
    def __init__(
            self,
            vpc: Vpc,
            subnets: List[Subnet],
            security_groups: List[SecurityGroup],
    ) -> None:
        """
        Constructor.

        :param vpc: VPC your function should be deployed to.
        :param subnets: List of subnets your function should be deployed to.
        :param security_groups: List of security groups for your function.
        """
        self.vpc = vpc
        self.subnets = subnets
        self.security_groups = security_groups



from typing import Optional, List

from aws_ci_cd_lambda.parameters.ssh_parameters import SshParameters


class PipelineParameters:
    """
    Parameters, focused on retrieving an ssh key for remote repositories and other pipeline actions.
    """
    def __init__(
            self,
            ssh_params: SshParameters,
            install_args: Optional[List[str]] = None,
            test_args: Optional[List[str]] = None,
            custom_pre_build_commands: Optional[List[str]] = None
    ) -> None:
        """
        Constructor.

        :param ssh_params: Parameters, focused on retrieving an ssh key for remote repositories.
        :param install_args: Arguments for your ./install.sh script
        :param test_args: Arguments for your ./test.sh script
        :param custom_pre_build_commands: Commands, that CodeBuild should execute between installation and testing. Optional
        """

        self.ssh_params = ssh_params
        self.install_args = install_args
        self.test_args = test_args
        self.custom_pre_build_commands = custom_pre_build_commands



from typing import Optional, List
from aws_cdk.aws_s3 import Bucket


class BuildSpecObject:
    def __init__(
            self,
            prefix: str,
            bucket: Bucket,
            aws_secret_id: Optional[str] = None,
            ssh_key: Optional[str] = None,
            install_args: Optional[List[str]] = None,
            test_args: Optional[List[str]] = None,
            custom_pre_build_commands: Optional[List[str]] = None
    ) -> None:
        assert aws_secret_id is None or ssh_key is None, 'Both aws secret id and ssh key cannot be set. Choose one.'

        self.__prefix = prefix
        self.__bucket = bucket
        self.__aws_secret_id = aws_secret_id
        self.__private_key = ssh_key
        self.__install_args = install_args or []
        self.__test_args = test_args or []
        self.__custom_pre_build_commands = custom_pre_build_commands or []

    def get_object(self):
        if self.__aws_secret_id is not None:
            install_ssh_commands = [
                'apt install jq',
                '{ aws secretsmanager get-secret-value --secret-id ' + self.__aws_secret_id +
                '| jq --raw-output \'.SecretString\' > id_rsa;'
                'eval `ssh-agent`; mv id_rsa ~/.ssh;'
                'chmod 0600 ~/.ssh/id_rsa;'
                'ssh-add ~/.ssh/id_rsa; } || { echo \"Invalid key\"; }'
            ]
        elif self.__private_key is not None:
            install_ssh_commands = [
                'apt install jq',
                '{ echo ' + self.__private_key + ' > id_rsa;'
                'eval `ssh-agent`; mv id_rsa ~/.ssh;'
                'chmod 0600 ~/.ssh/id_rsa;'
                'ssh-add ~/.ssh/id_rsa; } || { echo \"Invalid key\"; }'
            ]
        else:
            install_ssh_commands = []

        return {
            'version': 0.2,
            'phases': {
                'install': {
                    'commands': install_ssh_commands + [
                        'VENV_PATH="/tmp/lambda-tmpenv"',
                        'virtualenv $VENV_PATH --python=python3.6',
                        '. $VENV_PATH/bin/activate',
                        'chmod +x install.sh',
                        ' '.join(['./install.sh'] + self.__install_args)
                    ]
                },
                'pre_build': {
                    'commands': self.__custom_pre_build_commands + [
                        'chmod +x test.sh',
                        ' '.join(['./test.sh'] + self.__test_args)
                    ]
                },
                'build': {
                    'commands': [
                        'INSTALL_PATH="/tmp/ivs-lambda-install-dir"',
                        'BUILD_PATH="/tmp/ivs-lambda-pack.zip"',
                        'cp -R . $INSTALL_PATH',
                        'cp -R $VENV_PATH/lib/python3.6/site-packages/. $INSTALL_PATH',
                        'cp -R $VENV_PATH/lib/python3.6/site-packages/. .',
                        'current_dir=$( pwd )',
                        'cd $INSTALL_PATH',
                        'zip -9 -r $BUILD_PATH .',
                        'cd $current_dir',
                        f'KEY={self.__prefix}',
                        f'aws s3 cp $BUILD_PATH s3://{self.__bucket.bucket_name}/"$KEY".zip',
                        f'aws lambda update-function-code --function-name="$KEY" '
                        f'--s3-bucket={self.__bucket.bucket_name} --s3-key="$KEY".zip --publish'
                    ]
                },
            }
        }



from typing import Optional, Tuple


class SshParameters:
    """
    Parameters, focused on retrieving an ssh key for remote repositories.
    """
    def __init__(
            self,
            aws_secret: Optional[Tuple[str, str, Optional[str]]] = None,
            ssh_key: Optional[str] = None,
    ) -> None:
        """
        Constructor. At most one parameter must be fulfilled, as they are all multiple ways of retrieving an SSH key.

        :param aws_secret: An aws secret id and aws secret arn pair which point to a secret from which
        an ssh key can be retrieved. The secret should be stored as PLAINTEXT. If the secret is encrypted using
        a KMS key, the third member should be the ARN of that key, otherwise the third member should be None. Read more:
        https://docs.aws.amazon.com/secretsmanager/latest/userguide/tutorials_basic.html
        :param kms_key_arn: If your secret is encrypted using a KMS key, you need to specify the ARN of that key,
        so CodeBuild can decrypt the secret using that key.
        :param ssh_key: A direct supply of an ssh key without needing to retrieve it from anywhere else.
        """
        message = 'Only one way of retrieving SSH key is allowed.'
        assert sum([arg is not None for arg in [aws_secret, ssh_key]]) in [1, 0], message

        self.secret_id = None
        self.secret_arn = None
        self.kms_key_arn = None

        if aws_secret:
            self.secret_id, self.secret_arn, self.kms_key_arn = aws_secret

        self.private_key = ssh_key
